/* Copyright (C) 2000 Rainer Wichmann                                      */
/*                                                                         */
/*  This program is free software; you can redistribute it                 */
/*  and/or modify                                                          */
/*  it under the terms of the GNU General Public License as                */
/*  published by                                                           */
/*  the Free Software Foundation; either version 2 of the License, or      */
/*  (at your option) any later version.                                    */
/*                                                                         */
/*  This program is distributed in the hope that it will be useful,        */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/*  GNU General Public License for more details.                           */
/*                                                                         */
/*  You should have received a copy of the GNU General Public License      */
/*  along with this program; if not, write to the Free Software            */
/*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.              */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <gtk/gtk.h>
#include "Cora.h"

/* flag for open log window
 */
static int log_is_open = 0;       

/* the log buffer
 */
#define LOG_LINES    256
#define LOG_LINE_MAX  256
static  char log_buffer[LOG_LINES][LOG_LINE_MAX];
static  int  log_lines   = 0;
static  int  log_pos     = 0;
static  int  log_updated = 0;

/* the log window
 */
static GtkWidget   *log_window;

/* the log text
 */
static GtkWidget   *log_text;


static void close_log ( GtkWidget *widget, gpointer fs )
{
  log_is_open = 0;
  gtk_widget_destroy (GTK_WIDGET (fs));
  return;
}

static void delete_log ( )
{
  /* Freeze the text widget, ready for multiple updates           */
  gtk_text_freeze (GTK_TEXT (log_text));

  /* delete everything                                            */
  gtk_text_backward_delete(GTK_TEXT (log_text),
                           gtk_text_get_length( GTK_TEXT (log_text)));


  /* Thaw the text widget, allowing the updates to become visible */
  gtk_text_thaw (GTK_TEXT (log_text));

  return;
}

static void WriteLogText ()
{

   GdkFont *fixed_font;         /* the fixed font                 */
   GdkFont *italic_font;        /* the italic font                */ 
   GdkFont *normal_font;        /* the normal font                */ 
   GdkFont *bold_font;          /* the bold font                  */

   int         bold   = 0;      /* the bold flag                  */
#if 0
   int         italic = 0;      /* the italic flag                */
   int         fixed  = 0;      /* the fixed flag                 */
#endif
   char      * inStrg;          /* the string to put in viewer    */

   int         i;               /* standard counter variable      */
   int         pos;             /* log text position              */

   if (log_updated == 1)
     return;

   if (log_is_open == 0)
     return;

   delete_log();

  /* ------ Freeze the text widget, ready for multiple updates -- */

  gtk_text_freeze (GTK_TEXT (log_text));


  /* Load  fonts                                                  */
  fixed_font  = gdk_font_load 
               ("-misc-fixed-medium-r-*-*-*-120-*-*-*-*-*-*");
  bold_font   = gdk_font_load 
               ("*-new century schoolbook-bold-r-normal--*-120-*-*-*-*-*-*");
  italic_font = gdk_font_load 
               ("*-new century schoolbook-bold-i-normal--*-120-*-*-*-*-*-*");
  normal_font = gdk_font_load 
               ("*-new century schoolbook-medium-r-normal--*-120-*-*-*-*-*-*");


  /* the first line */
  i = 0;

  while (1) {

    /* LOG_LINES is maximum, log_lines is actual number of lines
     */
    if (i >= LOG_LINES || i >= log_lines)
      break;

    if (log_lines >= LOG_LINES)
      {
	pos = i + log_pos;
    
	if (pos >= LOG_LINES)
	  pos = pos % LOG_LINES;
      }
    else 
      pos = i;

    inStrg = log_buffer[pos];

    /* not special, print                                         */
    if (inStrg[0] != '*')
      {
	if (inStrg[0] == '#') inStrg[0] = ' ';
	gtk_text_insert (GTK_TEXT (log_text), normal_font, 
			 &log_text->style->black, NULL,
			 inStrg, -1);
	++i;
	continue;
      }
    

    /* special, set flag                                          */
    if (inStrg[0] == '*') bold   = 1;


    /* tagged word - bold                                         */
    if (bold == 1)
      {
	gtk_text_insert (GTK_TEXT (log_text), bold_font, 
			 &log_text->style->black, NULL,
			 inStrg, -1);
	++i;
	bold = 0;
	continue;
      }

  }


  /* Thaw the text widget, allowing the updates to become visible */  

  gtk_text_thaw (GTK_TEXT (log_text));

  log_updated = 1;

  return;

}

void insert_in_buffer (char * data)
{
  char * p;
  int    i;

  p = strtok(data, "\n");

  while (p || (p != '\0'))
    {
      strncpy (log_buffer[log_pos], p, LOG_LINE_MAX - 2);
      i = strlen(log_buffer[log_pos]);
      log_buffer[log_pos][i]   = '\n';
      log_buffer[log_pos][i+1] = '\0';

      ++log_pos;
      if (log_pos >= LOG_LINES)
	log_pos = 0;

      ++log_lines;
      if (log_lines > LOG_LINES)
	log_lines = LOG_LINES;

      p = strtok(NULL, "\n");
    }
 
  /* the buffer is modified, thus the display is out of date
   */
  log_updated = 0;

  /* update the display
   */
  if (log_is_open)
    WriteLogText();

  return;
}


void log_event ( /* GtkWidget *widget, gpointer *data  */)
{

  GtkWidget *hbox;
  GtkWidget *vbox;
  GtkWidget *button;
  GtkWidget *vscrollbar;


  /* ------- check whether already open  ----------------------   */

  if (log_is_open == 1) 
    {
      if (log_updated == 1)
	return;
      else
	{
	  WriteLogText ();
	}
      
      return;
    }

  /* ----------- open log viewer  -----------------------------   */


  log_window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_widget_set_usize (log_window, 580, 400);
  gtk_window_set_policy (GTK_WINDOW(log_window), TRUE, TRUE, FALSE);  
  gtk_signal_connect (GTK_OBJECT (log_window), "destroy",
                        (GtkSignalFunc) close_log, 
                         GTK_WIDGET (log_window));
  gtk_window_set_title (GTK_WINDOW (log_window), "Log");
  gtk_container_border_width (GTK_CONTAINER (log_window), 0);


  vbox = gtk_vbox_new (FALSE, 0);
  gtk_container_add (GTK_CONTAINER (log_window), vbox);
  gtk_widget_show (vbox);

  hbox = gtk_hbox_new (FALSE, 0);
  gtk_box_pack_start (GTK_BOX (vbox), hbox, TRUE, TRUE, 0);
  gtk_widget_show (hbox);

  /* ---------  Create the GtkText widget ----------------------  */

  log_text = gtk_text_new (NULL, NULL);
  gtk_text_set_editable (GTK_TEXT (log_text), FALSE);
  gtk_box_pack_start (GTK_BOX (hbox), log_text, TRUE, TRUE, 0);
  gtk_widget_show (log_text);
  
  /* ------- Add a vertical scrollbar to the GtkText widget ----- */

  vscrollbar = gtk_vscrollbar_new (GTK_TEXT (log_text)->vadj);
  gtk_box_pack_start(GTK_BOX(hbox), vscrollbar, FALSE, FALSE, 0);
  gtk_widget_show (vscrollbar);

  /* ----------- create buttons and add ------------------------- */

  hbox = gtk_hbox_new (FALSE, 0);
  gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
  gtk_widget_show (hbox);

  button = gtk_button_new_with_label ("Close");

  gtk_signal_connect (GTK_OBJECT (button), "clicked",
                      (GtkSignalFunc) close_log,
                      GTK_WIDGET (log_window));
  gtk_box_pack_start (GTK_BOX (hbox), button, TRUE, TRUE, 0);
  (GTK_WIDGET_FLAGS (button)  |= (GTK_CAN_DEFAULT));
  gtk_widget_grab_default (button); 

  gtk_widget_show (button);

  /* -------- Realizing the widget to insert some text ---------- */

  gtk_widget_realize (log_text);

  /* ------------------- set flag   ----------------------------- */

  log_is_open = 1;

  /* ------------------- insert text ---------------------------- */

  WriteLogText();

  /* ------------- and show it ---------------------------------- */

  gtk_widget_show (log_window);

  return;
}

void  my_log_handler  (const gchar *log_domain,
		       GLogLevelFlags log_level,
		       const gchar *message,
		       gpointer user_data)
{
  char * msg = malloc (strlen(message) + 1 + 32);

  char bad[] = "** ERROR **: Out of memory\n";
  
  if (msg != NULL)
    {
      if       ( (log_level & G_LOG_LEVEL_WARNING) != 0)
	strcpy(msg, "** WARNING **: ");
      else if  ( (log_level & G_LOG_LEVEL_MESSAGE) != 0)
	strcpy(msg, "Message      : ");
      strcat (msg, message);
      insert_in_buffer (msg);
    }
  else
    {
      insert_in_buffer (bad);
    }

  if (!quiet)
    {
      if (msg != NULL)
	fputs(msg, stderr);
      else
	fputs(message, stderr);
      fputc('\n', stderr);
    }

  if (msg != NULL)
    free(msg);
  
  return;
}
