#include "cora_fit.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include "cpgplot.h"

char datafile[MAXPATH];
char conffile[MAXPATH];
char proffile[MAXPATH];
double dt=1.;

/*   vectors read from the spectrum   */
Vector wave;    /* wavelength vector    */
Vector sp;      /* count vector         */
Vector err;     /* count errors         */
Vector bkgd;    /* instr. background    */
Vector bkerr;	/* error in background	*/
Vector bg0;     /* original background  */
double exposure=1.;	/* exposure time from datafile */
double binsize; /* binsize in Angstroem */

/*   vectors restricted to the selected part of the spectrum  */
Vector good;   /* indices of considered wavelength interval */
int    count;	/* number of indices in good  */
double *yc[MAXLINES];    /* matrix with nlines rows*/
Vector x;
Vector dx;
Vector ncount;
Vector bbb;
Vector bb0;

/* ------  fit results ----------------------- */
/*                                             */
Vector llambda;        /* lambda               */
Vector erlam;          /* error for lambda     */
Vector lsigma;         /* sigma                */
Vector ersig;          /* error for sigma      */
Vector ampout;         /* amplitude            */
Vector eramp;          /* error for amplitude  */

Vector steplam;        /* line distances to be kept constant */
Vector stepsig;	       /* distance between sigmas */
Vector f;	       /* array with weight factors for auxiliary conditions
                        * in the form sum(f_i*ampout_i)=v */
double v;

Vector lP;            /* lower borders for all variables        */
Vector mP;            /* start vector for iteration with Powell */
Vector uP;            /* upper borders for all variables        */
Vector startP;        /* start vector for all variables         */
double (*funcname)();  /* name of functions used by Powell    */
void (*proffunc)();  /* name of profile functions used in like.c */

double lamstart;   /* lower bound of investigated wavelength interval*/
double lamend;     /* upper bound of investigated wavelength interval*/
double lsig;       /* lower value for sigma */
double usig;       /* upper value for sigma */

/*  input parameters */
char fittitle[MAXTEXTL];
double dlam;	   /* search interval for lambda */
int    nlines;     /* number of lines   */
int    idlam;      /* switch for constant distance of lines
                                (-> steplam=const) */
int    idsig;      /* switch for same sigma for all lines */
double app_s;      /* apperture of source image */
double app_b;      /* apperture of backgr image */
int fixflag=1;     /* switch decides whether badly binned spectra will be
 fixed ( empty bins filles with zeros) */
double dsig;       /* search interval for sigma */
int    sigunit;    /* units of sigma: =1 -> km/s else Angstroem */
int    ikev;       /* unit of wavelength in plot: =1 in keV else A */
char   *linID;     /* label of line */
int    ilamit;     /* switch for iteration of lambda */
double sbg;        /* fixed value of source background if not iterated */
int    isigit;     /* switch for iteration of sigma */
int    sbgit;      /* switch for iteration of source background */
int    bgsmooth;   /* switch for smoothing of instrumental backgr */
int    sbgchi;	   /* switch for iteration of sbg with chi^2 */
double medsig=3.;     /* allowed interval for median background */ 
int    bg_dg;	   /* degree of polylom for background smoothing  */
int    iformat;    /* switch for specifying the output format*/
int    iportrait;  /* switch for orientation (=1 is portrait)  */
int    ires=0;     /* =1 print results into file resfile */
int    spec_only=0; /* =1 plot only the spectrum; no fitting */
int    fit_only=0;  /* =1 plot only the fit, no spectrum */
int    iplot;      /* =1 create plot of fit */
int    ilab;       /* labels will be insterted into plot
                                at adequate positions */
int    ierrplot;   /* switch for drawing error bars in plot */
int    ilog;       /* switch for logarithmic plot */
int    nbin;       /* number of bins beeing combined */
int    icount;     /* =1: counts will be conserved  
                      =0: flux will be conserved */
int    igraph=0;   /* >0 lines can be selected from spectrum
			(calling option -g) */
double iord;	   /* factor to shift multiplet, e.g. for higher order */
int    nlik;       /* switch for plotting liklihood curves
                                (also number of bins) */
int    ispline;    /* switch for fitting liklihood curves
                                with cubic splines */
int    iuse_fixpt[MAXLINES];/* switch to use fixed point equation */

int    verbose = 0;/* controls verbosity */

int    iprof=0;    /* flag for eventual numerical profile =1   */
int    pro_type[MAXLINES];        /* the profile type          */
float  eta_voig[MAXLINES];        /* the mixing parameter      */
int    fit_wave[MAXLINES];        /* whether to fit wavelength */
int    fit_sigm[MAXLINES];        /* whether to fit sigme      */
int    fit_lock[MAXLINES];        /* lock lambda to previous   */
float  beta[MAXLINES];             /* exponent in Beta model    */
int    imaxfit;                   /* currently unused, but set */
int    iget_final_result = FALSE; /* TRUE: switch off Voigt    */

/*    <<<<<<<<<   Main Program  >>>>>>>>>>>>   */
int main (int argc, char *argv[])

{

/*   parameters and variables
     variable names of input data comprehensive for the program

	wave	: wavelength vector [Angstoem]
	sp	: source count vector [cts]
	err	: vector with errors [cts]
	bkgd	: vector with instrumental + source background [cts]
		  the source background is added only for the run
		  of this program;
		  bkgd will be returned in the original form.
	date_s  : date of source exposure (string)
	date_b  : date of background exposure (string)
	fname1	: name of data file (string)

 --------------------------------------------
     other variable names used in the program
	ampout  : actual value of amplitude
	app_s/_b: appertures of source/backgr image
	bbb     : smoothed background in search interval
	bg0     : vector with weighted original instrumental background [cts]
	bweight : weighting factor for background from different apperturas
	eramp   : error in amplitude
	erlam   : error in lambda
	ersig   : error in sigma
	fname1  : name of datafile (will be printed into the plot)
	fname3  : name of input file
	good    : index vector of actual analyzed range
	idlam   : switch for constant distance of lines (-> steplam=const)
	idsig   : switch for same sigma for all lines
	ierrplot: switch for drawing error bars in plot
	iinput  : =0 read from COMMON block instead of input file (for autofit)
	ilab    : labels will be insterted into plot at adequate positions
	ilamit  : switch for iteration of lambda
	iformat : switch for specifying the output format:
		  0 - terminal (default)
		  1 - PS file
		  2 - GIF file
		  3 - PPM file
	iportrait: switch for orientation (=1 is portrait)
	iord    : investigate higher order of same line
                done by multiplication of starting value(s) of lambda with iord
	iplot   : =1 create plot of fit
	ilog	: switch for logarithmic plot
	ires    : =1 print results results into file resfile
	isigit  : switch for for iteration in sigma
	ispline : switch for fitting liklihood curves with cubic splines
	lamend  : upper limits of lambda
	lamstart: lower limits of lambda
	linID   : label of line
	llambda : actual value of best lambda
	lsig    : lower value for sigma
        usig    : upper value for sigma
	sbg	: constant source background (continuum)
	sbgchi	: switch for iteration of sbg with chi^2 method
	medsig  : allowed interval for median background
	sigunit : units of sigma: =1 -> km/s else Angstroem
	lsigma  : actual value of best sigma
	nbin    : number of bins beeing combined (integral will be conserved)
	ncount  : count vector in search interval
	nlik    : switch for plotting liklihood curves (also number of bins)
	nlines  : number of lines
	num_sig : parameter for width of interval
		  within which likelihood is determined (cf. fname3)
	sbg     : estimated continuum background of source
	sbgit   : iterate sbg
	steplam	: array with linedistances which can be forced to be constant
	tol     : tolerance parameter for Brents and Powells iteration
	unit    : channel number for reading input file
	x       : wavelength vector in search interval good
	dx	: vector with distances of bin (averaged)
	yc      : vector containing theoretical curve of lines
   */


  int    i, lline, flag=1, nb;
  FILE * fd;
  Vector sp_b,dummy;
  double sbgmed,sbgA,bweight,like;
  char   resfile[MAXPATH], usage[MAXPATH], plotfile[MAXPATH];
  char su[MAXTEXTL], au[MAXTEXTL]="cts";
  float  ampstore[MAXLINES];

  int ndigit=0,j=0,vk;
  float test;
  char *fformat;
  time_t time_total, time_slice;

  /* initialize to default: gauss profile, fit lambda
   */
  for(i=0; i < MAXLINES; ++i)
   {
    fit_wave[i] = TRUE;
    fit_sigm[i] = TRUE;
    fit_lock[i] = FALSE;
    pro_type[i] = P_GAUSS;
    eta_voig[i] = 1.0;
    beta[i] = 1.0;
   }

  vec_init(&wave);
  vec_init(&sp);
  vec_init(&err);
  vec_init(&bkgd);
  vec_init(&bkerr);
  vec_init(&bg0);
  vec_init(&good);
  vec_init(&llambda);
  vec_init(&erlam);
  vec_init(&lsigma);
  vec_init(&ersig);
  vec_init(&ampout);
  vec_init(&eramp);
  vec_init(&steplam);
  vec_init(&stepsig);
  vec_init(&f);
  vec_init(&lP);
  vec_init(&mP);
  vec_init(&uP);
  vec_init(&startP);
  vec_init(&x);
  vec_init(&dx);
  vec_init(&ncount);
  vec_init(&bbb);
  vec_init(&bb0);

  vec_init(&sp_b);
  vec_init(&dummy);

  strncpy (datafile, "spectrum.dat", MAXPATH);
  strncpy (conffile, "fit.inp", MAXPATH);
  strncpy (proffile, "profile.dat", MAXPATH);
  strncpy (resfile, "cora.dat", MAXPATH);
  strncpy (plotfile, "cora", MAXPATH);
  strncpy (usage, "Usage: cora_fit [-help] [-s spectrum_file] [-i parameter_file] "\
	   "[(-o|-r) result_file] [-p plot_file] [-e expos. time] [-l lineprofile_file] -v -v -g\n", 
	   MAXPATH);

  while (argc > 1 && argv[1][0] == '-')
    {
      switch (argv[1][1])
        {
	  /*************
	case 'a':
	  iuse_fixpt = 0;
	  break; 
	  ***************/
	case 'V':
	  fprintf (stderr, "Cora line fitting tool Version %s\n",VERSION);
	  exit (8);
	case 'v':
	  ++verbose;
	  break;
        case 's':
          ++argv;
          --argc;
          if (argc == 1)
            {
              fprintf (stderr,usage); 
              exit (8);
            }
          strncpy (datafile, argv[1], MAXPATH);
          break;
        case 'l':
          ++argv;
          --argc;
          if (argc == 1)
            {
              fprintf (stderr,usage);
              exit (8);
            }
          strncpy (proffile, argv[1], MAXPATH);
	  iprof = 1;
          break;
        case 'i':
          ++argv;
          --argc;
          if (argc == 1)
            {
              fprintf (stderr, usage);
              exit (8);
            }
          strncpy (conffile, argv[1], MAXPATH);
          break;
	case 'o':
	  ++argv;
	  --argc;
	  if (argc == 1)
	    {
	      fprintf (stderr, usage);
	      exit (8);
	    }
	  strncpy (resfile, argv[1], MAXPATH);
	  break;
	case 'p':
	  ++argv;
	  --argc;
	  if (argc == 1)
	    {
		fprintf (stderr, usage);
		exit (8);
	    }
	  strncpy (plotfile, argv[1], MAXPATH);
	  break;
        case 'r':  
	  ++argv;
	  --argc;
	  if (argc == 1)
	    {
		fprintf (stderr,usage);
		exit (8);
	    }
	  strncpy (resfile, argv[1], MAXPATH);
	  if(replot(resfile,plotfile)==0)
	    printf("Message      : Executed replot with %s successfully\n;;\n",
                                 resfile);
	  else break;
	  return(0);
          break;
	case 'e':
	  ++argv;
	  --argc;
	  if (argc == 1)
	    {
	      fprintf (stderr, usage);
	      exit (8);
	    }
	  dt = atof(argv[1]);
	  if (dt != 1.)
	   printf("* Warning *: The exposure time will take effect in the log file, which should not be used for replots\n");
	  break;
	case 'g':
	  ++igraph;
	  break;
	default:
	  fprintf (stderr, usage);
          fprintf (stderr, "%s","spectrum file    : spectrum with two, three, or five columns (create example with cora_spec)\n");
          fprintf (stderr, "%s","parameter file   : Input file (create default with cora_inp)\n");
          fprintf (stderr, "%s","result_file      : file containing fit results:\n");
          fprintf (stderr, "%s","  -o             : indicate file to be written into\n");
          fprintf (stderr, "%s","  -r             : indicate file to be read from (for replot)\n");
          fprintf (stderr, "%s","plot_file        : (ps, gif, or ppm) file\n");
          fprintf (stderr, "%s","expos. time      : exposure time (can also be given in first line of spectrum file)\n");
          fprintf (stderr, "%s","lineprofile_file : file with numeric profile function (two columns)\n");
          fprintf (stderr, "%s","-v               : verbosity (up to three times)\n");
          fprintf (stderr, "%s","-g               : Open graph for interactive slection of lines\n");
	  exit (8);
	  break;
        }
      ++argv;
      --argc;
    }

  if (NULL == getenv("PGPLOT_DIR") && NULL == getenv("PGPLOT_FONT"))
    {
      fprintf(stderr, 
	      "* Warning *: Neither the environment variable PGPLOT_DIR\n"\
	      "*            nor the environment variable PGPLOT_FONT\n"\
	      "*            are defined.\n"\
	      "*            This means that PGPLOT will not be able to\n"\
	      "*            locate the binary font file 'grfont.dat', and\n"\
	      "*            thus plot labels may be corrupted or missing.\n"\
	      "*            Please consult your system administrator for\n"\
	      "*            proper setting of (at least) one of these two\n"\
	      "*            environment variables.\n");
    }

  /*  read data file  
   */
  if (verbose)
    fprintf(stdout, "Message      : Reading spectrum   from %s...\n", 
	    datafile);
  if (0 >= read_spec (datafile, 0))
    {
      fprintf(stderr, 
	      "* Warning *: No spectrum data read in.\n");
      exit (3);
    }
/*  binsize = (wave.vec[wave.len-1] - wave.vec[0]) / (wave.len-1);*/ 
    vec_alloc(&dummy,wave.len-1);
    for (i=1;i<wave.len;++i) dummy.vec[i-1]=wave.vec[i]-wave.vec[i-1];
    binsize = median(&dummy);
    vec_free(&dummy);

  /*  read parameter file  
   */
  if (verbose)
    fprintf(stdout, "Message      : Reading parameters from %s...\n", 
	    conffile);
  if (0 >= read_input (conffile, 0))
    {
      fprintf(stderr,
              "* Warning *: No parameter file read in.\n");
      exit (3);
    }

  nb = check_bins();

  /* icount = 1: counts (rather than flux) will be conserved while rebinning
   */
  icount = 1;


  /* determine line distances to be kept constant if idlam = 1,
   * and flag all except first line as fit_wave = FALSE  
   */
  for (lline = 0; lline < (nlines-1); ++lline)
    {
      steplam.vec[lline] = llambda.vec[lline+1] - llambda.vec[lline];
      stepsig.vec[lline] = lsigma.vec[lline+1]  - lsigma.vec[lline];
    }
/*  if (idlam == 1)
    {
      for (lline = 1; lline < nlines; ++lline)
	fit_wave[lline] = FALSE;
      fit_wave[0] = TRUE;
    }
 */

  if (idsig > 0)
    {
      i=0;
      while (lsigma.vec[i] <= 0.) ++i;
      for (lline = i+1; lline < nlines; ++lline)
        if (fit_sigm[lline] == TRUE) lsigma.vec[lline] = lsigma.vec[i];
    }

  /* set flags to FALSE if isigit == 0 || ilamit == 0
   */
  if (isigit == 0)
    {
      for (lline = 1; lline < nlines; ++lline)
	fit_sigm[lline] = FALSE;
    }
  if (ilamit == 0)
    {
      for (lline = 1; lline < nlines; ++lline)
	fit_wave[lline] = FALSE;
    }


  /* determine dataintervall to be investigated  
   */
  if (verbose)
    printf("Message      : Determining wavelength interval under "\
	   "investigation\n");
  lamran();
  if (verbose>1)
    printf("Message      : ... [%f ... %f] A\n", 
	   wave.vec[(int)good.vec[0]], wave.vec[(int)good.vec[count-1]]);


  /* 
   * count is the number of bins in the spectral range of interest,
   * and is set in lamran()
   */


  /* allocate vector to hold spectral background
   */
  vec_alloc(&sp_b, count);

  /*  allow for different appertures of source and background exposure  
   */
  bweight = app_s/app_b;

  /* define two background vectors:
	bg0  : pure instrumental background
	bkgd : total background, i.e. SUM(bg0_i+sbg)
   */
  for (i=0; i<count; ++i) 
    {
      bg0.vec[(int)good.vec[i]]  = bg0.vec[(int)good.vec[i]]  * bweight;
      bkgd.vec[(int)good.vec[i]] = bkgd.vec[(int)good.vec[i]] * bweight;
      bkerr.vec[(int)good.vec[i]] = bkerr.vec[(int)good.vec[i]] * bweight;
    }


  /*   define current environment, i.e. extract the spectral range of
   *   interest to us
   */
  for (i = 0; i < count; ++i)
    {
      x.vec[i]      = wave.vec[(int)good.vec[i]]; /* wavelenghts */
      ncount.vec[i] =   sp.vec[(int)good.vec[i]]; /* raw counts  */
      bbb.vec[i]    = bkgd.vec[(int)good.vec[i]]; /* instr. bg   */
      bb0.vec[i]    =  bg0.vec[(int)good.vec[i]]; /* total bg    */
    }
    for (i = 1; i < count-1; ++i) dx.vec[i]=0.5*(x.vec[i+1]-x.vec[i-1]);
    dx.vec[0] = x.vec[1]-x.vec[0];
    dx.vec[count-1] = x.vec[count-1]-x.vec[count-2];

  /* choose proper likelihood function wrappers
      and set iteration limits
   */
  if (verbose)
   {
     printf("Message      : Setting valid domain for lambda and sigma\n");
     if (verbose >1)
       printf("Message      :    with dlam=%f  ;   dsig=%f\n",dlam,dsig);
   }
  borders();

  /*  smooth background vector bkgd with m-order polonomial 
   */
  if (bgsmooth > 0) smoothbg();

  /* define profile procedure
   */
  if (iprof==1) proffunc = num_profile; else proffunc = profile;

  /*****************************************************
   *  
   * TREATMENT OF CONTINUUM
   *  
   *****************************************************/

  /* determine source background by median function and/or chi^2 approach 
   */
 sbgA=sbg;
 sbg = sbg*binsize; /*(wave.vec[wave.len-1]-wave.vec[0])/wave.len;*/
  if (sbgit > 0 || sbgchi > 0)
    {
      for (i=0; i<count; ++i)
        sp_b.vec[i] = ncount.vec[i]-bbb.vec[i];
      sbgmed = median(&sp_b);
      sbg    = sbgiterate(&sp_b,sbgmed);
      vec_free(&sp_b);
    }
  if (sbg < 0)
    sbg = 0;
  if (verbose>1)
    printf("Message      : sbg = %f cts/bin\n",sbg);
  sbgA = sbg/binsize; /*(wave.vec[wave.len-1]-wave.vec[0])*wave.len;*/

  /*  add source background to data interval under consideration 
   */
  for (i=0; i < count; ++i) 
    bbb.vec[i] = bbb.vec[i]+sbg;

  /*****************************************************
   *
   * PLOT LIKELIHOOD CURVE
   *
   *****************************************************/

if (nlik > 0)
  {
  fprintf(stderr, "Message      : nlik>0: Will create likelihood curves\n");
  fprintf(stderr, "        This feature is currently only available for the first line\n");
test=llambda.vec[0];
   plotlik_lam(0);
llambda.vec[0]=test;
   plotlik_sig(0);
   exit(8);
  }

  /*****************************************************
   *
   * START THE FIT
   *
   *****************************************************/
/*
for (lline = 0; lline < nlines; ++lline)
        fprintf(stderr,"%f\n",lsigma.vec[lline]);
for (lline = 0;lline<mP.len;++lline)
        fprintf(stderr,"%d %f\n",mP.len,mP.vec[lline]);
*/
  time_slice = time(&time_total);

  if (verbose)
    {
      if (ilamit+isigit > 0) 
	fprintf(stdout, "Message      : starting fit\n");
      else 
	fprintf(stdout, "Message      : calculate Amplitudes\n");
    }

  if (ilamit+isigit > 0) 
    powell(&(mP.vec[0]), mP.len, 1.0e-6, &like, funcname, 0.1);
  else 
    like = funcname(&(mP.vec[0]), &mP.len);

  if (verbose)
    {
      if (ilamit+isigit > 0)
        fprintf(stdout, "Message      : end fit (%ld sec)\n", 
	        (long) (time(NULL) - time_slice));
      else
        fprintf(stdout, "Message      : end calculate Amplitudes (%ld sec)\n",
                (long) (time(NULL) - time_slice));
    }

  /*****************************************************
   *
   * START THE ERROR ANALYSIS
   *
   *****************************************************/

  time_slice = time(NULL);

  if (verbose)
    fprintf(stdout, "Message      : starting error analysis (lambda/sigma)\n");


  if (like != LIKE_PENALTY)
    {
      
      para_err(); 

      if (verbose)
	fprintf(stdout, "Message      : finished error analysis (%ld sec)\n",
                 (long) (time(NULL) - time_slice)); 
    }
  else
    fprintf(stderr,"* Warning *: iteration ran out of allowed range\n");

  /*  checkerr();  */


  /*   get final result -- must compute BEFORE Amperr() to get
   *   individual gauss/lorentz profile functions for voigt profile 
   */
  if (verbose)
    fprintf(stdout, "Message      : calculating last likelihood\n");

  like = funcname(&(mP.vec[0]), &mP.len); 

  if (verbose)
    fprintf(stdout, "Message      : likelihood -2 ln(L) = %8.4f\n", like);



  /* iget_final_result = TRUE  will switch off Voigt profile, to get
   * individual profile functions from Gaussian and Lorentzian part.
   */
  for (i = 0; i < nlines; ++i)
    ampstore[i] = ampout.vec[i];
   
  iget_final_result = TRUE;
  (void) funcname(&(mP.vec[0]), &mP.len); 
  iget_final_result = FALSE;

  for (i = 0; i < nlines; ++i)
    {
      if (ampstore[i] == 0.0 && eta_voig[i] != 0.0 && i < (nlines-1) && f.vec[i] == 0)
	ampout.vec[i] = eta_voig[i] * ampstore[i+1];
      else 
	ampout.vec[i] = eta_voig[i] * ampstore[i];
    }

  if (like != LIKE_PENALTY)
    {
      if (verbose)
        fprintf(stdout, 
                "Message      : starting error analysis (amplitudes)\n");
      time_slice = time(NULL);
      Amperr(&ampout);
      if (verbose) 
        fprintf(stdout, "Message      : finished error analysis (%ld sec)\n",
                (long) (time(NULL) - time_slice));
    } 

  
  for (i = 0; i < nlines; ++i)
    if (!finite(eramp.vec[i])) 
      eramp.vec[i] = -1.0;
  
  /*  write log file
   */
  if (dt != 1.) sprintf(au,"%s","cts/s");
  if (ires > 0)
    {
      /* check whether log file already exists
       */
      if (verbose) printf("Message      : writing log file\n");
      fd = fopen (resfile, "r");
      if(!fd)
	flag=1;
      else
	{
	  flag=0;
	  fclose(fd);
	}

/*  determine correct format for amlitudes rates */
     fformat= malloc(59);
     test=99999999999999.;
     for (i=1;i<nlines;++i) if (eramp.vec[i] <test &&eramp.vec[i] != 0.) {test=eramp.vec[i];j=i;}
     ndigit = floor((log10(eramp.vec[j]/dt)));
     if (ndigit<0) {ndigit=-ndigit+2;vk=ndigit+4;}
     else {vk = ndigit+3;ndigit=3;}

     if (dt != 1.)
	sprintf(fformat,"%s8.3f  %s8.5f  %s7.4f  %s7.5f  %s%d.%df  %s%d.%df  %s%d.%df %sd\n","%","%","%","%","%",vk,ndigit,"%",vk,ndigit,"%",vk,ndigit,"%");
     else
	sprintf(fformat,"%s8.3f  %s8.5f  %s7.4f  %s7.5f  %s10.4f  %s7.4f   %s7.2f  %sd\n","%","%","%","%","%","%","%","%");
 
      fd = fopen (resfile, "a");
      
      /* write header only if file is new
       */
      if(flag==1)
	fprintf(fd,
		"# lambda[A] +- err   Sigma[A] +- err      A[%s] +- err"\
		"    sbg[%s/A]\n",au,au);
      for(i=0;i<nlines;++i)
	fprintf(fd,fformat,
		llambda.vec[i], erlam.vec[i],
		lsigma.vec[i],  ersig.vec[i],
		ampout.vec[i]/dt,  eramp.vec[i]/dt,
		sbgA/dt,
		pro_type[i]);
      fclose  (fd);
    }
  
  /*  create backup file of last fit result
   */
  
  if (0 == test_confdir())
    {
      strncpy(resfile, get_home_dir(), MAXPATH);
      strncat(resfile, "/.cora/cora.dat", MAXPATH-strlen(resfile)-1);
      if (verbose) printf("Message      : writing backup file\n");
      fd = fopen (resfile, "w+");
      for(i=0;i<nlines;++i)
	fprintf(fd,
		"%8.3f  %8.5f  %7.4f  %7.5f  %10.4f  %7.4f   %7.2f  %d\n",
		llambda.vec[i], erlam.vec[i],
		lsigma.vec[i],  ersig.vec[i],
		ampout.vec[i],  eramp.vec[i],
		sbgA,
		pro_type[i]);
      fclose  (fd);
    }
  
  /*  convert sigma back to desired units for output
       (formerly changed in read_input into A);
      Angstroem are used for internal purpose
   */
  if (sigunit == 1)
    {
      if (verbose) printf("Message      : converting sigmas to units km/s\n");
      dsig = C_LIGHT*dsig/llambda.vec[0];
      for(i=0;i<nlines;++i)
	{
	  lsigma.vec[i]=C_LIGHT*lsigma.vec[i]/llambda.vec[i];
	  ersig.vec[i]=C_LIGHT*ersig.vec[i]/llambda.vec[i];
	}
    }
  
  
  /*  output of final results on screen
   */

  if (verbose)
    fprintf(stdout, "Message      : finished  (%ld sec)\n",
            (long) (time(NULL) - time_total));

  
  fprintf(stdout, ";\n");
  fprintf(stdout, "; RESULT:\n");
  fprintf(stdout, "; =======\n");
  fprintf(stdout, ";  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n");
  fprintf(stdout, ";   %s\n",fittitle);
  fprintf(stdout, ";  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n");
if (sigunit == 0)
  sprintf(su,"%s","Angstroem");
else
  sprintf(su,"%s","km/s");

    fprintf(stdout, 
	    "; Lambda [Angstroem]    Sigma [%s]    Amplitude [%s]\n",su,au);
  
  for(i=0;i<nlines;++i) 
    {
     if (erlam.vec[i] > 0. && ersig.vec[i] > 0.)
      {
       fprintf(stdout, 
	      "; %7.4f +- %7.4f   %7.4f +- %7.4f   %7.2f +- %7.4f  %s\n", 
	      llambda.vec[i], erlam.vec[i],
	      lsigma.vec[i],  ersig.vec[i],
	      ampout.vec[i]/dt,  eramp.vec[i]/dt,
	      (pro_type[i] == P_GAUSS) ? "GAUSS" : "LORENTZ");
      
       fprintf(stdout,
	      ";   [= %7.5f +- %7.4f keV]  [FWHM=%6.4f +- %6.4f A]\n",
	      12.398544/llambda.vec[i], 
	      12.398544*erlam.vec[i]/(llambda.vec[i]*llambda.vec[i]),
	      (pro_type[i] == P_GAUSS) ? 2.35482*lsigma.vec[i] : 
	      2 * lsigma.vec[i],
	      (pro_type[i] == P_GAUSS) ? 2.35482*ersig.vec[i] :
	      2 * lsigma.vec[i]);
      } else
       {
        if (erlam.vec[i] > 0. && ersig.vec[i] <= 0.)
	{
           fprintf(stdout,
              "; %7.4f +- %7.4f   %7.4f  NoErr     %7.2f +- %7.4f  %s\n",
              llambda.vec[i], erlam.vec[i],
              lsigma.vec[i],
              ampout.vec[i]/dt,  eramp.vec[i]/dt,
              (pro_type[i] == P_GAUSS) ? "GAUSS" : "LORENTZ");

           fprintf(stdout,
              ";   [= %7.5f +- %7.4f keV]  [FWHM=%6.4f A]\n",
              12.398544/llambda.vec[i],
              12.398544*erlam.vec[i]/(llambda.vec[i]*llambda.vec[i]),
              (pro_type[i] == P_GAUSS) ? 2.35482*lsigma.vec[i] :
              2 * lsigma.vec[i]);
	}
	if (erlam.vec[i] <= 0. && ersig.vec[i] > 0.)
	{
           fprintf(stdout,
              "; %7.4f    NoErr     %7.4f +- %7.4f   %7.2f +- %7.4f  %s\n",
              llambda.vec[i],
              lsigma.vec[i],  ersig.vec[i],
              ampout.vec[i]/dt,  eramp.vec[i]/dt,
              (pro_type[i] == P_GAUSS) ? "GAUSS" : "LORENTZ");

           fprintf(stdout,
              ";   [= %7.5f keV]  [FWHM=%6.4f +- %6.4f A]\n",
              12.398544/llambda.vec[i],
              (pro_type[i] == P_GAUSS) ? 2.35482*lsigma.vec[i] :
              2 * lsigma.vec[i],
              (pro_type[i] == P_GAUSS) ? 2.35482*ersig.vec[i] :
              2 * lsigma.vec[i]);
	}
	if (erlam.vec[i] <= 0. && ersig.vec[i] <= 0)
	{
           fprintf(stdout,
              "; %7.4f    NoErr     %7.4f     NoErr  %7.2f +- %7.4f  %s\n",
              llambda.vec[i],
              lsigma.vec[i],  
              ampout.vec[i]/dt,  eramp.vec[i]/dt,
              (pro_type[i] == P_GAUSS) ? "GAUSS" : "LORENTZ");

           fprintf(stdout,
              ";   [= %7.5f keV]  [FWHM=%6.4f A]\n",
              12.398544/llambda.vec[i],
              (pro_type[i] == P_GAUSS) ? 2.35482*lsigma.vec[i] :
              2 * lsigma.vec[i]);
        }
       }
    }
  
  fprintf(stdout, "; sbg    = %8.4f cts/Angstroem \n", sbgA);
  fprintf(stdout, "; like   = %8.4f\n", like);
  fprintf(stdout, "; ========================\n");
  fprintf(stdout, ";\n");
  if (min(&erlam)<=0.0 || min(&ersig)<=0.0)
    {
      fprintf(stdout, "; Attention:\n");
      fprintf(stdout, "; ==========\n");
      fprintf(stdout, "; Errors with NoErr mean that the "\
	      "variable is not iterated or\n");
      fprintf(stdout, "; that the variable was not iterated "\
	      "beyond its iteration limits.\n");
    }
  
  /*
   *  ** WARNING **: double semicolon signals EOF 
   */
  fprintf(stdout, ";;\n");
  fflush(stdout);
  
  
  /*   create nice plot
   */
  if(iplot == 1)
    plotfit (conffile       /* title             */,
	     plotfile       /* filename if device PS, GIF or PPM */,
	     iformat        /* XSERVE/PS/GIF/PPM */
	     );
  
  vec_free(&dummy);
  vec_free(&sp_b);
  
  return(0);
}
